--
-- PLB component for EMS
--  320x200x8 VGA driver
--  LEDs and switches - local & remote access
--  Debugger
--  User UART
--  User memory access
-- 
-- Copyright (C) Jack Whitham 2009
-- http://www.jwhitham.org.uk/
--
-- 

library ieee;
library unisim ;

use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned."+";
use ieee.std_logic_unsigned."-";
use ieee.std_logic_unsigned.conv_integer;
use unisim.vcomponents.all ;


entity plbvga_core is
    generic (
        ADD_DEBUGGER        : Boolean := true;
        ADD_USER_UART       : Boolean := true;
        ADD_USER_MEMORY_IF  : Boolean := true;
        CPU_LEFT            : Natural := 31;
        DATA_LEFT           : Natural := 31);
    port (
-- External connections
        -- Video output
        red                 : out std_logic;
        green               : out std_logic;
        blue                : out std_logic;
        hsync               : out std_logic;
        vsync               : out std_logic;

        -- Serial ports
        ser_txd			    : out std_logic;        -- User's serial port
        ser_rxd			    : in std_logic;         -- (M14/R7 - DCE)
        debug_serial_out    : out std_logic;        -- Debug serial port
        debug_serial_in     : in std_logic;         -- (U8/M13 - DTE)

        -- Wires to/from simple hardware
        leds_out            : out std_logic_vector ( 7 downto 0 );
        switches_in         : in std_logic_vector ( 3 downto 0 );
        btn_north_in        : in std_logic;
        btn_east_in         : in std_logic;
        btn_south_in        : in std_logic;
        btn_west_in         : in std_logic;
        rot_centre_in       : in std_logic;
        rot_a               : in std_logic;
        rot_b               : in std_logic;


-- User interfaces - connect to Handel C
        -- User interface for memory
        mem_user_address    : in std_logic_vector ( 31 downto 0 );
        mem_data_in         : in std_logic_vector ( 31 downto 0 );
        mem_data_out        : out std_logic_vector ( 31 downto 0 );
        mem_read            : in std_logic;
        mem_write           : in std_logic;
        mem_busy            : out std_logic;
        mem_byte_enable     : in std_logic_vector ( 3 downto 0 );
        hc_rst              : out std_logic;

        -- User UART channel (in/out)
        uart_data_in        : in std_logic_vector ( 7 downto 0 );
        uart_data_out       : out std_logic_vector ( 7 downto 0 );
        uart_read           : in std_logic;
        uart_write          : in std_logic;
        uart_data_present   : out std_logic;
        uart_busy           : out std_logic;

        -- User connections for simpler hardware
        btn_north_out       : out std_logic;
        btn_east_out        : out std_logic;
        btn_south_out       : out std_logic;
        btn_west_out        : out std_logic;
        rot_centre_out      : out std_logic;
        rot_ccw             : out std_logic;
        rot_cw              : out std_logic;
        leds_in             : in std_logic_vector ( 7 downto 0 );
        switches_out        : out std_logic_vector ( 3 downto 0 );

-- Internal connections 
        -- PLB signals
        control_in          : in std_logic_vector ( 31 downto 0 );
        control_wrce        : in std_logic_vector ( 0 to 3 );
        dbg_out             : out std_logic_vector ( 31 downto 0 );

        -- Control lines from plbvga to IPIF
        mst_cntl_rd_req     : out std_logic ;
        mst_cntl_wr_req     : out std_logic ;
        mst_cntl_bus_lock   : out std_logic ;
        mst_cntl_burst      : out std_logic ;
        mst_ip2bus_addr     : out std_logic_vector ( CPU_LEFT downto 0 ) ;
        mst_ip2bus_be       : out std_logic_vector ( 15 downto 0 ) ;
        mst_xfer_length     : out std_logic_vector ( 11 downto 0 ) ;
        mst_go              : out std_logic ;

        -- Status lines from IPIF to plbvga
        mst_cmd_sm_set_done : in std_logic ;
        mst_cmd_sm_set_error : in std_logic ;
        mst_cmd_sm_set_timeout : in std_logic ;
        mst_cmd_sm_busy     : in std_logic ;
        mst_cmd_sm_clr_go   : in std_logic ;

        -- Data interface
        output_next         : in std_logic ;
        input_next          : in std_logic ;
        data_output         : out std_logic_vector ( DATA_LEFT downto 0 ) ;
        data_input          : in std_logic_vector ( DATA_LEFT downto 0 ) ;
        -- Clock
        clk_100             : in std_logic ;
        clk_50              : in std_logic ;
        reset               : in std_logic ) ;
end entity plbvga_core ;


architecture basic of plbvga_core is

-- Here are some possible video modes that you can use.
-- Other resolutions are possible but require a different ratio of
-- input pixels to output pixels
-- "1024x768"    65     1024 1032 1176 1344    768  771  777  806 -hsync -vsync
-- "1024x768"    62     1024 1064 1240 1280   768  774  776  808
-- "1024x768"    75     1024 1048 1184 1328    768  771  777  806 -hsync -vsync
-- "1024x768"    72     1024 1056 1192 1280    768  770  776 806   -hsync -vsync
-- "1024x768"    85     1024 1032 1152 1360    768  784  787  823

    -- These control the ratio of input pixels (320x200) to output pixels
    constant X_MAX                  : Natural := 2; 
    constant Y_MAX                  : Natural := 2;
    constant SCREEN_VISIBLE_WIDTH   : Natural := 960; -- (1+X_MAX) * 320
    constant SCREEN_HEIGHT          : Natural := 600; -- (1+Y_MAX) * 200

    -- These control the horizontal and vertical scan
    constant SCREEN_SCAN_WIDTH      : Natural := 1360;
    constant SCREEN_SCAN_HEIGHT     : Natural := 823; 
    constant SCAN_X_HBI_START       : Natural := 1032 + 32; -- 32=(1024-960)/2
    constant SCAN_X_HBI_END         : Natural := 1152 + 32;
    constant SCAN_Y_VBI_START       : Natural := 784 - 84; -- 84=(768-600)/2
    constant SCAN_Y_VBI_END         : Natural := 787 - 84;

    constant INVERT_HSYNC           : std_logic := '0'; -- this is -hsync
    constant INVERT_VSYNC           : std_logic := '0'; -- this is -vsync


    -- Uart configuration
    constant DEBUG_BAUD     : Natural := 568181;
    constant USER_BAUD      : Natural := 115200;
    constant TIC_RATE       : Natural := 35 * 4;

    constant DEBUG_BAUD_COUNTER_MAX : Natural := 100e6 / ( DEBUG_BAUD * 16 );
    constant USER_BAUD_COUNTER_MAX : Natural := 50e6 / ( USER_BAUD * 16 );
    constant TC_MAX         : Natural := ( USER_BAUD * 16 ) / TIC_RATE ;

    signal tic_counter      : std_logic_vector ( 9 downto 0 );
    signal tic_counter_interval: Natural range 0 to TC_MAX;


    -- Bus types
    subtype Bus_CPU_Address is std_logic_vector ( CPU_LEFT downto 0 ) ;
    subtype Bus_CPU_Data is std_logic_vector ( DATA_LEFT downto 0 ) ;
    subtype Palette_Level is std_logic_vector ( 5 downto 0 ) ;

    -- The following control the fetching of data from memory
    constant PACKETS_PER_LINE       : Natural := 5;
    constant BITS_PER_PIXEL         : Natural := 8;
    constant PACKET_SIZE            : Natural := 16;
    constant NUMBER_OF_LINES        : Natural := SCREEN_HEIGHT;
    constant PIXELS_PER_WORD        : Natural := 32 / BITS_PER_PIXEL;
    constant MEMORY_LINES           : Natural := 200;
    constant WORDS_DRAWN_PER_LINE   : Natural := 80;
    constant WORDS_FETCHED_PER_LINE : Natural := 80;
    constant BURSTS_REQUIRED        : Natural := 
                                            MEMORY_LINES * PACKETS_PER_LINE;

    signal start_address    : std_logic_vector ( CPU_LEFT downto 0 );
    signal palette_input    : std_logic_vector ( DATA_LEFT downto 0 );

    signal trigger, fault_reg: std_logic; 
    signal mst_go_int       : std_logic;
    signal busy, fault      : std_logic;
    signal video_reset      : std_logic;
    signal vsync_active     : std_logic;
    signal hsync_active     : std_logic;
    signal drawing_active   : std_logic;
    signal some_words_remain: std_logic;
    signal one_word_remains : std_logic;
    signal memory_operation : std_logic;
    signal fifo_wr          : std_logic;
    signal fifo_rd          : std_logic;
    signal fifo_rd_1        : std_logic;
    signal fifo_below_lower_threshold : std_logic;
    signal fifo_above_upper_threshold : std_logic;

    -- Video
    signal scan_x           : std_logic_vector ( 10 downto 0 ) ;
    signal scan_y           : std_logic_vector ( 10 downto 0 ) ;
    signal pixel_out        : std_logic_vector ( BITS_PER_PIXEL - 1 downto 0 );
    signal draw_x, draw_y   : std_logic;
    signal next_pixel       : std_logic;
    signal subscan_x        : std_logic_vector ( 2 downto 0 ) ;

    -- Additional output register stage
    signal d_red, d_green   : std_logic;
    signal d_blue, d_hsync  : std_logic;
    signal d_vsync          : std_logic;

    -- Internal registers for memory access
    signal words_remaining  : Natural range 0 to PACKET_SIZE ;
    signal mem_remaining    : Natural range 0 to BURSTS_REQUIRED;
    signal mem_address      : Bus_CPU_Address ;

    -- Bayer diffusion
    signal bayer_address, store_address : std_logic_vector(9 downto 0);
    signal red_level, green_level       : Palette_Level;
    signal blue_level, blue_value       : Palette_Level;
    signal red_value, green_value       : Palette_Level;
    signal store_input, store_output    : std_logic_vector(15 downto 0);
    signal bayer_out                    : std_logic_vector(15 downto 0);
    signal x_address                    : std_logic_vector(9 downto 0);
    signal store_write                  : std_logic;

    subtype Cycle_X_Type is Natural range 0 to X_MAX;
    subtype Cycle_Y_Type is Natural range 0 to Y_MAX;

    constant WRITE_PHASE    : Cycle_X_Type := 1;
    constant PALETTE_UPDATE : Cycle_X_Type := 2;
    constant READ_COMPLETE  : Cycle_X_Type := 1;
    constant X_ADVANCE      : Cycle_X_Type := 2; -- mark
    constant Y_FIRST        : Cycle_Y_Type := 0;
    signal cycle_x          : Cycle_X_Type;
    signal cycle_y          : Cycle_Y_Type;

    -- Debugger
    type Debug_State_Type is ( DBG_INITIAL, DBG_WAIT_LEDS, DBG_NEW_PIXEL,
                DBG_WAIT_ECHO, DBG_GET_FIFO_PIXEL, DBG_AWAIT_FIFO, DBG_STOP,
                DBG_LOAD, DBG_SEND_BYTE, DBG_SPOOL, 
                DBG_ESCAPED, DBG_WAIT_START);
    type Debug_Send_Type is ( SEND_NOTHING, SEND_START,
                SEND_PIXEL, SEND_ECHO, SEND_LEDS, SEND_RUN_LENGTH );
    signal debug_state              : Debug_State_Type ;
    signal debug_active             : std_logic;
    signal debug_send               : Debug_Send_Type;
    signal debug_data_sent          : std_logic;
    signal send_character           : std_logic;
    signal send_character_reg       : std_logic;
    signal tx_complete              : std_logic;
    signal debug_command_ready      : std_logic;
    signal debug_command_out        : std_logic_vector(7 downto 0);
    signal debug_out                : std_logic_vector(7 downto 0);
    signal debug_en_16_x_baud       : std_logic;

    signal fifo_empty               : std_logic;
    signal no_data                  : std_logic;

    signal last_pixel_out           : std_logic_vector(7 downto 0);
    signal run_length               : std_logic_vector(6 downto 0);

    signal debug_baud_counter       : Natural range 0 to DEBUG_BAUD_COUNTER_MAX;

    constant MAX_REQUEST            : Natural := 3;
    signal request_echo             : Natural range 0 to MAX_REQUEST;
    signal request_frame            : Natural range 0 to MAX_REQUEST;
    signal echo_char                : std_logic;

    -- User things - rotation decoder
    signal rot_rotation             : std_logic;
    signal rot_rotation_d           : std_logic;
    signal rot_is_ccw               : std_logic;
    signal rot_a_ff                 : std_logic;
    signal rot_b_ff                 : std_logic;
    signal rot_in                   : std_logic_vector ( 1 downto 0 );
    signal button_ready             : std_logic;
    signal button_trigger           : std_logic_vector(2 downto 0);

    signal debounce_ready           : std_logic_vector(4 downto 0);
    signal debounce_out             : std_logic_vector(4 downto 0);
    signal debounce_in              : std_logic_vector(4 downto 0);

    -- User things - UART
    signal useruart_baud_counter    : Natural range 0 to USER_BAUD_COUNTER_MAX;
    signal useruart_write           : std_logic;
    signal useruart_sc              : std_logic;
    signal useruart_data_strobe     : std_logic;
    signal useruart_done            : std_logic;
    signal useruart_present         : std_logic;
    signal useruart_data            : std_logic_vector(7 downto 0);
    signal useruart_en_16_x_baud    : std_logic;

    -- User memory access
    signal mem_done                 : std_logic;
    signal mem_read_latch           : std_logic;
    signal mem_write_latch          : std_logic;

    -- Memory access state machine
    type State_Type is ( 
            VBI_RESET, STANDBY, FETCH_DATA, STAGE3, 
            USER_ACCESS, SLOW_ACCESS_WAIT ) ;
    
    signal state            : State_Type ;

attribute INIT_00 : string ;
attribute INIT_00 of palette_and_line_store : label is
     "0ca710c814e90041006204a40cc50000042108420c637fff2529002204430000";
attribute INIT_01 : string ;
attribute INIT_01 of palette_and_line_store : label is
     "1cf41cf42115253629572d77319835b939da3dfb421b463c4a5d529e56be5adf";
attribute INIT_02 : string ;
attribute INIT_02 of palette_and_line_store : label is
     "000800080009000a000b000b042c042d084e084e0c6f0c70109114b114b218d3";
attribute INIT_03 : string ;
attribute INIT_03 of palette_and_line_store : label is
     "25f92a192e3a325b367c3a9d3ebe42df4aff4eff531f5b3f5f5f637f6b9f6fbf";
attribute INIT_04 : string ;
attribute INIT_04 of palette_and_line_store : label is
     "048508a608a70cc90cea110b110d152e154f195019711d931db421b521d625f7";
attribute INIT_05 : string ;
attribute INIT_05 of palette_and_line_store : label is
     "46314a524a524e73529456b55ad65ad65ef7631867396b5a6f7b6f7b739c77bd";
attribute INIT_06 : string ;
attribute INIT_06 of palette_and_line_store : label is
     "1084108414a518c61ce7210821082529294a2d6b318c35ad35ad39ce3def4210";
attribute INIT_07 : string ;
attribute INIT_07 of palette_and_line_store : label is
     "0041048204c209030d44118515c616071a481e6922aa26eb2b2b2f6c33ad37ee";
attribute INIT_08 : string ;
attribute INIT_08 of palette_and_line_store : label is
     "14ea190a192b1d4c214d256e258f29b02dd131f235f336133a343e5542764697";
attribute INIT_09 : string ;
attribute INIT_09 of palette_and_line_store : label is
     "10e6150719281d4a218b25ac29cd31ef0cc810e9150b194c1d6e25b029d13213";
attribute INIT_0A : string ;
attribute INIT_0A of palette_and_line_store : label is
     "0c7f1cff2d7f3dff4e7f5eff6f7f7fff00ae011009730df5167822fa2b7d3bff";
attribute INIT_0B : string ;
attribute INIT_0B of palette_and_line_store : label is
     "00080009000b000c000e000f001100130014001600170019001a001c001d001f";
attribute INIT_0C : string ;
attribute INIT_0C of palette_and_line_store : label is
     "2800340040004c005800640070007c007c637cc67d4a7dce7e317eb57f187f9c";
attribute INIT_0D : string ;
attribute INIT_0D of palette_and_line_store : label is
     "0115011601380159057a059b05bd09de0dff1e3f2e9f3edf4f1f5f5f6fbf7fff";
attribute INIT_0E : string ;
attribute INIT_0E of palette_and_line_store : label is
     "046508860ca810e9009000b200d300f403ff13ff23ff37ff47ff5bff6bff7fff";
attribute INIT_0F : string ;
attribute INIT_0F of palette_and_line_store : label is
     "35b4340d4c1364197c1f7dff279f227f000004000800100014001c0020002800";
attribute INIT_10 : string ;
attribute INIT_10 of palette_and_line_store : label is
     "00100018000a0012000f001700090011001f0008001a0002001f000700190001";
attribute INIT_11 : string ;
attribute INIT_11 of palette_and_line_store : label is
     "000c0014000e0016000b0013000d0015001c0004001e0006001b0003001d0005";
attribute INIT_12 : string ;
attribute INIT_12 of palette_and_line_store : label is
     "000f00170009001100100018000a0012001f000700190001001f0008001a0002";
attribute INIT_13 : string ;
attribute INIT_13 of palette_and_line_store : label is
     "000b0013000d0015000c0014000e0016001b0003001d0005001c0004001e0006";

begin 

    -- Internal PLB controls
    busy <= mst_cmd_sm_busy or mst_go_int or trigger ;
    some_words_remain <= '1' when ( words_remaining /= 0 )
                else '0' ;
    one_word_remains <= '1' when ( words_remaining = 1 )
                else '0' ;

    -- Outgoing controls to PLB IPIF
    mst_cntl_bus_lock <= '0' ;
    mst_go <= mst_go_int ;
    mst_xfer_length <= "000000000100"       -- 1 word
                when (ADD_USER_MEMORY_IF and
                    ((state = USER_ACCESS) or (state = SLOW_ACCESS_WAIT)))
            else "000001000000";            -- 16 words
    mst_cntl_burst <= memory_operation
            and some_words_remain and not one_word_remains ;
    mst_ip2bus_addr <= mem_user_address
            when (ADD_USER_MEMORY_IF and
                ((state = USER_ACCESS) or (state = SLOW_ACCESS_WAIT)))
            else mem_address;
    mst_ip2bus_be <= mem_byte_enable & mem_byte_enable & 
                    mem_byte_enable & mem_byte_enable ; 
    data_output <= mem_data_in;
    fault <= mst_cmd_sm_set_timeout or mst_cmd_sm_set_error or fault_reg;

    -- FIFO buffer 
    vf : entity vgafifo 
        port map (
            rd_clk => clk_100,
            wr_clk => clk_50,
            din => data_input,
            rd_en => fifo_rd,
            rst => video_reset,
            wr_en => fifo_wr,
            dout => pixel_out,
            empty => fifo_empty,
            full => open,
            prog_empty => fifo_below_lower_threshold,
            prog_full => fifo_above_upper_threshold);


    fifo_wr <= 
        '1' when ((input_next = '1') and (some_words_remain = '1')
                and (state = STAGE3))
            else '0';

    no_data <= 
        '1' when start_address (31 downto 28) = x"0" else '0';

    mem_done <= 
        '1' when (((busy = '0') or (fault = '1'))
            and (state = SLOW_ACCESS_WAIT)) else '0';

    -- State machine for memory system
    process ( clk_50 , reset ) is
    begin
        if ( reset = '1' )
        then
            state <= VBI_RESET;
            mst_go_int <= '0' ;
            mst_cntl_rd_req <= '0';
            mst_cntl_wr_req <= '0';
            fault_reg <= '0';

        elsif ( clk_50'event )
        and ( clk_50 = '1' )
        then
            trigger <= '0' ;
            fault_reg <= fault;

            case state is
            when VBI_RESET =>
                    -- This state is held during the vertical blanking
                    -- interval. No memory operations during that
                    -- time.
                    mst_cntl_rd_req <= '0';
                    mst_cntl_wr_req <= '0';
                    fault_reg <= '0';
                    mem_address <= start_address;
                    mem_remaining <= BURSTS_REQUIRED;
                    if (((mem_read_latch = '1') or (mem_write_latch = '1'))
                    and ADD_USER_MEMORY_IF)
                    then
                        state <= USER_ACCESS; 
                    elsif ((video_reset = '0')
                    and (no_data = '0'))
                    then
                        state <= FETCH_DATA;
                    end if;
        
            when STANDBY =>
                    -- This state is held when there is sufficient data
                    -- for drawing within the FIFO. 
                    mst_cntl_rd_req <= '0';
                    mst_cntl_wr_req <= '0';
                    fault_reg <= '0';
                    if (fifo_below_lower_threshold = '1')
                    then
                        if (mem_remaining /= 0)
                        then
                            state <= FETCH_DATA;
                        end if;
                    elsif (((mem_read_latch = '1') or (mem_write_latch = '1'))
                    and ADD_USER_MEMORY_IF)
                    then
                        state <= USER_ACCESS; 
                    end if;
                    if (video_reset = '1')
                    then
                        state <= VBI_RESET;
                    end if;

            when FETCH_DATA =>
                    words_remaining <= PACKET_SIZE;
                    mst_cntl_rd_req <= '0';
                    mst_cntl_wr_req <= '0';
                    if (video_reset = '1')
                    then
                        state <= VBI_RESET;
                    elsif ((fifo_above_upper_threshold = '1')
                    or (mem_remaining = 0))
                    then
                        state <= STANDBY;
                    else
                        memory_operation <= '1' ;
                        mst_cntl_rd_req <= '1';
                        trigger <= '1' ;
                        state <= STAGE3 ;
                    end if;

            when STAGE3 =>
                    if ( fault = '1' )
                    then
                        state <= VBI_RESET;
                    elsif ( busy = '0' )
                    then
                        mst_cntl_rd_req <= '0';
                        mem_address (31 downto 6) <= 
                                mem_address (31 downto 6) + 1 ;
                        mem_remaining <= mem_remaining - 1;
                        state <= FETCH_DATA;
                    end if ;

                    if (( input_next = '1' )
                    and ( some_words_remain = '1' ))
                    then
                        words_remaining <= words_remaining - 1 ;
                    end if ;

            when USER_ACCESS =>
                    state <= SLOW_ACCESS_WAIT;
                    words_remaining <= 1;
                    mst_cntl_rd_req <= not mem_write_latch;
                    mst_cntl_wr_req <= mem_write_latch;
                    trigger <= '1' ;

            when SLOW_ACCESS_WAIT =>
                    if ( input_next = '1' )
                    then
                        mem_data_out <= data_input ;
                    end if ;
                    if (( busy = '0' )
                    or ( fault = '1' ))
                    then
                        if (video_reset = '1')
                        then
                            state <= VBI_RESET;
                        else
                            state <= STANDBY;
                        end if;
                    end if ;

            end case ;

            -- Additional control lines
            if ( trigger = '1' )
            then
                mst_go_int <= '1' ;
            end if ;
            if ( mst_cmd_sm_clr_go = '1' )
            then
                mst_go_int <= '0' ;
            end if ;

            mem_address ( 1 downto 0 ) <= "00";
        end if;
    end process;

    -- State machine for VGA output

    process ( clk_100 , reset ) is
    begin
        if ( reset = '1' )
        then
            vsync_active <= '1';
            hsync_active <= '0';
            video_reset <= '1';
            scan_y <= conv_std_logic_vector(SCAN_Y_VBI_START, 11);
            scan_x <= ( others => '0' );
            subscan_x <= ( others => '0' );
            draw_x <= '0';
            draw_y <= '0';
            hc_rst <= '1';
            cycle_x <= 0;
            cycle_y <= 0;
            debug_send <= SEND_NOTHING;
            debug_state <= DBG_INITIAL;
            run_length <= (others => '0');

        elsif ( clk_100'event )
        and ( clk_100 = '1' )
        then
            fifo_rd_1 <= '0';
            debug_send <= SEND_NOTHING;

            if ((debug_active = '1')
            and (ADD_DEBUGGER))
            then
                video_reset <= '0';
                vsync_active <= '0';
                hsync_active <= '0';
                draw_x <= '0';
                draw_y <= '0';

                case debug_state is
                when DBG_INITIAL =>
                    run_length <= (others => '1');
                    video_reset <= '1';
                    scan_x <= (others => '0');
                    subscan_x <= (others => '0');
                    scan_y <= (others => '0');
                    cycle_x <= 0;
                    cycle_y <= 0;
                    if (request_echo /= 0)
                    then
                        debug_send <= SEND_ECHO;
                        debug_state <= DBG_WAIT_ECHO;
                    elsif (request_frame /= 0)
                    then
                        debug_send <= SEND_START;
                        debug_state <= DBG_WAIT_START;
                    end if;
                when DBG_WAIT_ECHO =>               
                    video_reset <= '1';
                    if (debug_data_sent = '1')
                    then
                        debug_state <= DBG_INITIAL;
                    end if;
                when DBG_WAIT_START =>               
                    video_reset <= '1';
                    if (debug_data_sent = '1')
                    then
                        debug_send <= SEND_LEDS;
                        debug_state <= DBG_WAIT_LEDS;
                    end if;
                when DBG_WAIT_LEDS =>
                    if (debug_data_sent = '1')
                    then
                        debug_state <= DBG_GET_FIFO_PIXEL;
                    end if;
                when DBG_GET_FIFO_PIXEL =>
                    last_pixel_out <= pixel_out;
                    if ((fifo_empty = '0') or (no_data = '1'))
                    then
                        fifo_rd_1 <= '1';
                        debug_state <= DBG_AWAIT_FIFO;
                    end if;
                when DBG_AWAIT_FIFO =>
                    debug_state <= DBG_LOAD;
                when DBG_LOAD =>
                    debug_state <= DBG_SEND_BYTE;
                when DBG_SEND_BYTE =>
                    if ((pixel_out = last_pixel_out)
                    and (run_length /= "1111111"))
                    then
                        -- Pixel is the same as its predecessor
                        -- Increment run length
                        run_length <= run_length + 1;
                        debug_state <= DBG_SPOOL;
                    else
                        debug_state <= DBG_NEW_PIXEL;
                    end if;
                when DBG_NEW_PIXEL|DBG_STOP =>
                    -- Pixel is not the same as its predecessor
                    -- Or the run_length is at maximum
                    if ((last_pixel_out(7) = '1')
                    or (debug_state = DBG_STOP)
                    or (run_length /= "0000000"))
                    then
                        -- Must emit the run length
                        debug_send <= SEND_RUN_LENGTH;
                        debug_state <= DBG_ESCAPED;
                    else
                        -- Just send the pixel
                        debug_state <= DBG_SPOOL;
                        debug_send <= SEND_PIXEL;
                    end if;
                when DBG_ESCAPED =>
                    if (debug_data_sent = '1')
                    then
                        -- Sent run length, now send the pixel
                        debug_send <= SEND_PIXEL;
                        debug_state <= DBG_SPOOL;
                        run_length <= (others => '0');
                    end if;
                when DBG_SPOOL =>
                    if (debug_data_sent = '0')
                    then
                        null; -- not sent yet
                    elsif (scan_x = conv_std_logic_vector(319, 11))
                    then
                        -- next pixel is on the next line
                        scan_y <= scan_y + 1;
                        if (scan_y = conv_std_logic_vector(199, 11))
                        then
                            -- There is no "next line".
                            -- Send out the final pixel
                            last_pixel_out <= pixel_out;
                            debug_state <= DBG_STOP;
                        elsif (scan_y = conv_std_logic_vector(200, 11))
                        then
                            -- This is the end.
                            debug_state <= DBG_INITIAL;
                        else
                            scan_x <= (others => '0');
                            debug_state <= DBG_GET_FIFO_PIXEL;
                        end if;
                    else
                        -- next pixel
                        scan_x <= scan_x + 1;
                        debug_state <= DBG_GET_FIFO_PIXEL;
                    end if;
                end case;
            else
                -- Not debugging. Use regular raster scan.
                debug_state <= DBG_INITIAL;

                if (cycle_x /= X_MAX)
                then
                    cycle_x <= cycle_x + 1;
                else
                    cycle_x <= 0;
                    scan_x <= scan_x + 1;
                end if;
                subscan_x <= subscan_x + 1;

                if (scan_x = conv_std_logic_vector(
                            (SCREEN_VISIBLE_WIDTH / (X_MAX + 1)), 11))
                then
                    draw_x <= '0';
                elsif (scan_x = conv_std_logic_vector(
                                SCAN_X_HBI_START/(X_MAX + 1), 11))
                then
                    hsync_active <= '1';
                elsif (scan_x = conv_std_logic_vector(
                                SCAN_X_HBI_END/(X_MAX + 1), 11))
                then
                    hsync_active <= '0';
                elsif (scan_x = conv_std_logic_vector(
                                SCREEN_SCAN_WIDTH/(X_MAX + 1), 11))
                then
                    scan_x <= ( others => '0');
                    scan_y <= scan_y + 1;
                    subscan_x <= ( others => '0');

                    cycle_x <= 0;
                    if (cycle_y /= Y_MAX)
                    then
                        cycle_y <= cycle_y + 1;
                    else
                        cycle_y <= 0;
                    end if;

                    if (scan_y = conv_std_logic_vector(
                                        SCREEN_SCAN_HEIGHT, 11))
                    then
                        scan_y <= ( others => '0' );
                        cycle_y <= 0;
                        draw_y <= '1';
                    elsif (scan_y = conv_std_logic_vector(
                                        SCREEN_SCAN_HEIGHT - 2, 11))
                    then
                        fifo_rd_1 <= '1';
                    elsif (scan_y = conv_std_logic_vector(
                                        SCREEN_HEIGHT - 1, 11))
                    then
                        draw_y <= '0';
                        hc_rst <= '0';
                    elsif (scan_y = conv_std_logic_vector(
                                        SCAN_Y_VBI_START, 11))
                    then
                        vsync_active <= '1';
                        video_reset <= '1';
                    elsif (scan_y = conv_std_logic_vector(
                                        SCAN_Y_VBI_END, 11))
                    then
                        vsync_active <= '0';
                        video_reset <= '0';
                    end if;

                    -- For the next clock cycle:
                    draw_x <= '1';
                end if;
            end if;
            scan_y ( 10 ) <= '0';
            scan_x ( 10 downto 9 ) <= "00";
        end if;
    end process;

    fifo_rd <= '1' when ((cycle_x = X_ADVANCE)
            and (cycle_y = Y_FIRST)
            and (debug_active = '0')
            and (drawing_active = '1'))
            else fifo_rd_1;


    x_address <= "1" & scan_x ( 8 downto 0 );

    store_address <= ("00" & palette_input(23 downto 16)) 
                    when (cycle_x = PALETTE_UPDATE)
        else x_address when ((cycle_y /= Y_FIRST)
                        or (cycle_x = WRITE_PHASE))
        else ("00" & pixel_out);
    store_write <= '1' when (((cycle_x = PALETTE_UPDATE))
                        or ((cycle_y = Y_FIRST)
                            and (cycle_x = WRITE_PHASE)))
        else '0'; 
    store_input(14 downto 0) <= 
            palette_input(14 downto 0) 
                    when (cycle_x = PALETTE_UPDATE)
        else store_output(14 downto 0);
    store_input(15) <= '1';
   
    palette_and_line_store : RAMB16_S18_S18
        port map ( 
            addrA => store_address, 
            diA => store_input,
            dipA => "00",
            doA => store_output,
            dopA => open,
            clkA => clk_100, 
            enA => '1', 
            weA => store_write,

            addrB => bayer_address,
            diB => x"0000",
            dipB => "00",
            doB => bayer_out,
            dopB => open,
            clkB => clk_100,
            enB => '1', 
            weB => '0',

            ssra => reset, ssrb => reset);

    bayer_address <= "0100" & 
            subscan_x & scan_y ( 2 downto 0 ) ;

    -- Output stage
    process ( clk_100 ) is
        variable bayer_value    : Palette_Level;
    begin
        if ( clk_100'event )
        and ( clk_100 = '1' )
        then
            if (cycle_x = READ_COMPLETE)
            then
                red_value <= (others => '0');
                green_value <= (others => '0');
                blue_value <= (others => '0');
                red_value (4 downto 0) <= store_output (4 downto 0) ;
                green_value (4 downto 0) <= store_output (9 downto 5) ;
                blue_value (4 downto 0) <= store_output (14 downto 10) ;

                drawing_active <= draw_x and draw_y;
            end if;

            bayer_value := (others => '0');
            bayer_value (4 downto 0) := bayer_out (4 downto 0) ;
            red_level <= bayer_value + red_value ;
            green_level <= bayer_value + green_value ;
            blue_level <= bayer_value + blue_value;
        end if;
    end process;

    process (drawing_active, red_level, green_level, blue_level,
            hsync_active, vsync_active) is
    begin
        d_red <= red_level(5);
        d_green <= green_level(5);
        d_blue <= blue_level(5);
        d_hsync <= hsync_active;
        d_vsync <= vsync_active;
        if (drawing_active = '0')
        then
            d_red <= '0';
            d_green <= '0';
            d_blue <= '0';
        end if;
    end process;

    red <= d_red;
    green <= d_green;
    blue <= d_blue;
    hsync <= d_hsync xor INVERT_HSYNC;
    vsync <= d_vsync xor INVERT_VSYNC;

    -- Serial interface for debugging
    d : if (ADD_DEBUGGER)
    generate
        krx : entity kcuart_rx 
            port map (
                    serial_in => debug_serial_in,
                    data_out => debug_command_out,
                    data_strobe => debug_command_ready,
                    en_16_x_baud => debug_en_16_x_baud,
                    clk => clk_100 );
        ktx : entity kcuart_tx 
            port map (
                    serial_out => debug_serial_out,
                    send_character => send_character,
                    en_16_x_baud => debug_en_16_x_baud,
                    data_in => debug_out,
                    Tx_complete => tx_complete,
                    clk => clk_100 );

        debug_en_16_x_baud <= '1' when (debug_baud_counter = 0) else '0';
    end generate;

    d2 : if (not ADD_DEBUGGER)
    generate
        debug_serial_out <= debug_serial_in;
        debug_command_out <= x"00";
    end generate;

    send_character <= send_character_reg and not tx_complete;
    debug_data_sent <= '1' when 
            ((send_character_reg = '0') and (debug_send = SEND_NOTHING))
            else '0';


    -- Counters for baud & tics
    dbg_out ( 7 downto 0 ) <= tic_counter(9 downto 2);
    process ( clk_100, reset ) is
    begin
        if ( reset = '1' )
        then
            debug_baud_counter <= 0;
            send_character_reg <= '0';
            button_trigger <= "111";
            request_frame <= 0;
            request_echo <= 0;
            echo_char <= '0';
            debug_active <= '0';

        elsif ( clk_100'event )
        and ( clk_100 = '1' )
        then
            if (ADD_DEBUGGER)
            then
                if (debug_baud_counter = 0)
                then
                    debug_baud_counter <= DEBUG_BAUD_COUNTER_MAX - 1;
                else
                    debug_baud_counter <= debug_baud_counter - 1;
                end if;

                case debug_send is
                when SEND_NOTHING =>
                        if (tx_complete = '1')
                        then
                            send_character_reg <= '0';
                        end if;
                when SEND_RUN_LENGTH =>
                        debug_out (6 downto 0) <= run_length;
                        debug_out (7) <= '1';
                        send_character_reg <= '1';
                when SEND_PIXEL =>
                        debug_out <= last_pixel_out;
                        send_character_reg <= '1';
                when SEND_ECHO =>
                        debug_out <= x"3" & "000" & echo_char;
                        request_echo <= request_echo - 1;
                        send_character_reg <= '1';
                        echo_char <= not echo_char;
                when SEND_START =>
                        debug_out <= x"53";
                        send_character_reg <= '1';
                when SEND_LEDS =>
                        debug_out <= leds_in;
                        request_frame <= request_frame - 1;
                        send_character_reg <= '1';
                end case ;

                if ((debug_command_ready = '1')
                and (debug_command_out (7 downto 5) /= "011"))
                then
                    debug_active <= '0';
                elsif (debug_command_ready = '1')
                then
                    debug_active <= '1';
                    button_trigger <= "111";
                    if (debug_command_out (4 downto 0) = "00110")
                    then
                        if (debug_send = SEND_LEDS)
                        then
                            request_frame <= request_frame;
                        elsif (request_frame /= MAX_REQUEST)
                        then
                            request_frame <= request_frame + 1;
                        end if;
                    elsif (debug_command_out (4 downto 0) = "00010")
                    then
                        if (debug_send = SEND_ECHO)
                        then
                            request_echo <= request_echo;
                        elsif (request_echo /= MAX_REQUEST)
                        then
                            request_echo <= request_echo + 1;
                        end if;
                    elsif (debug_command_out (4 downto 3) = "01")
                    then
                        button_trigger <= debug_command_out(2 downto 0);
                    elsif (debug_command_out (4) = '1')
                    then
                        switches_out <= debug_command_out(3 downto 0);
                    end if;
                end if;
            end if;
            if (debug_active = '0')
            then
                request_frame <= 0;
                request_echo <= 0;
                echo_char <= '0';
                switches_out <= switches_in;
            end if;
        end if;
    end process;

    -- Slave side PLB registers & user memory access
    process (reset, clk_50) is
    begin
        if ( reset = '1' ) 
        then
            start_address <= ( others => '0' ) ;
            palette_input <= ( others => '0' ) ;
            mem_write_latch <= '0';
            mem_read_latch <= '0';
        elsif ( clk_50'event and clk_50 = '1' ) 
        then
            if ( control_wrce(0) = '1') 
            then
                start_address <= control_in;
                start_address ( 1 downto 0 ) <= "00";
            end if;
            if ( control_wrce(1) = '1') 
            then
                palette_input <= control_in;
            end if;
            if (mem_write = '1')
            then
                mem_write_latch <= '1';
                mem_read_latch <= '0';
            elsif (mem_read = '1')
            then
                mem_write_latch <= '0';
                mem_read_latch <= '1';
            elsif (mem_done = '1')
            then
                mem_write_latch <= '0';
                mem_read_latch <= '0';
            end if;
        end if;
    end process;
    mem_busy <= mem_write or mem_read 
        or mem_write_latch or mem_read_latch;


    -- External things: LEDs, buttons and switches
    leds_out <= leds_in ;
    process ( clk_50 ) is
    begin
        if ( reset = '1' )
        then
            button_ready <= '1';
            debounce_out <= "00000";
            debounce_ready <= "00000";

        elsif ( clk_50 = '1' )
        and ( clk_50'event )
        then
            debounce_out <= "00000";

            if ((debug_active = '1')
            and ADD_DEBUGGER)
            then
                rot_rotation <= '0';
                button_ready <= '0';
                if (button_trigger = "111")
                then
                    button_ready <= '1';
                elsif (button_ready = '1')
                then
                    case button_trigger is
                    when "000" => debounce_out(0) <= '1';
                    when "001" => debounce_out(1) <= '1';
                    when "010" => debounce_out(2) <= '1';
                    when "011" => debounce_out(3) <= '1';
                    when "100" => debounce_out(4) <= '1';
                    when "101" => rot_is_ccw <= '1';
                                  rot_rotation <= '1';
                    when others =>rot_is_ccw <= '0';
                                  rot_rotation <= '1';
                    end case;
                end if;
            else
                button_ready <= '1';
                for i in 0 to 4
                loop
                    if (debounce_in(i) = '0')
                    then
                        debounce_ready(i) <= '1';
                    elsif ((debounce_ready(i) = '1')
                    and (tic_counter_interval = 0))
                    then
                        debounce_out(i) <= '1';
                        debounce_ready(i) <= '0';
                    end if;
                end loop;


                -- Rotary decoding
                case rot_in is
                when "00" =>    rot_rotation <= '0';         
                when "01" =>    rot_is_ccw <= '0';
                when "10" =>    rot_is_ccw <= '1';
                when "11" =>    rot_rotation <= '1';
                when others =>  null;
                end case;
            end if;

            -- Rotary switch filter
            rot_a_ff <= rot_a;
            rot_b_ff <= rot_b;
            rot_in <= rot_b_ff & rot_a_ff;

            rot_rotation_d <= rot_rotation;
            rot_ccw <= '0';
            rot_cw <= '0';

            -- Rotary switch output
            if (( rot_rotation = '1' )
            and ( rot_rotation_d = '0' ))
            then
                if ( rot_is_ccw = '1' )
                then
                    rot_ccw <= '1';
                else
                    rot_cw <= '1';
                end if;
            end if;
        end if;
    end process;

    debounce_in (0) <= btn_north_in;
    debounce_in (1) <= btn_east_in;
    debounce_in (2) <= btn_south_in;
    debounce_in (3) <= btn_west_in;
    debounce_in (4) <= rot_centre_in;
    btn_north_out <= debounce_out(0);
    btn_east_out <= debounce_out(1);
    btn_south_out <= debounce_out(2);
    btn_west_out <= debounce_out(3);
    rot_centre_out <= debounce_out(4);


    -- Serial interface for user
    u : if (ADD_USER_UART)
    generate
        ukrx : entity kcuart_rx 
            port map (
                    serial_in => ser_rxd,
                    data_out => useruart_data,
                    data_strobe => useruart_data_strobe,
                    en_16_x_baud => useruart_en_16_x_baud,
                    clk => clk_50 );
        uktx : entity kcuart_tx 
            port map (
                    serial_out => ser_txd ,
                    send_character => useruart_sc,
                    en_16_x_baud => useruart_en_16_x_baud,
                    data_in => uart_data_in,
                    Tx_complete => useruart_done,
                    clk => clk_50 );

        useruart_sc <= useruart_write and not useruart_done;
        uart_busy <= useruart_write or uart_write;
        uart_data_present <= useruart_present and not uart_read;

        process (clk_50, reset) is
        begin
            if (reset = '1')
            then
                useruart_write <= '0';
                uart_data_out <= x"00";
                useruart_baud_counter <= 0;
                tic_counter <= ( others => '0' ) ;
                tic_counter_interval <= 0;
            elsif ((clk_50 = '1')
            and (clk_50'event))
            then
                if (uart_write = '1')
                then
                    useruart_write <= '1';
                elsif (useruart_done = '1')
                then
                    useruart_write <= '0';
                end if;
                if (useruart_data_strobe = '1')
                then
                    uart_data_out <= useruart_data;
                    useruart_present <= '1';
                end if;
                if (uart_read = '1')
                then
                    useruart_present <= '0';
                end if;
                if (useruart_baud_counter = 0)
                then
                    if (tic_counter_interval = 0)
                    then
                        tic_counter_interval <= TC_MAX;
                        tic_counter <= tic_counter + 1;
                    else
                        tic_counter_interval <= tic_counter_interval - 1;
                    end if;
                    useruart_baud_counter <= USER_BAUD_COUNTER_MAX - 1;
                else
                    useruart_baud_counter <= useruart_baud_counter - 1;
                end if;
            end if;
        end process;
        useruart_en_16_x_baud <= '1' when (useruart_baud_counter = 0) else '0';

    end generate;

    u2 : if (not ADD_USER_UART)
    generate
        ser_txd <= ser_rxd;
    end generate;

    -- Features todo (and test):
    -- Single word/halfword/byte read/write support (retrofit?)
    -- Palette readback?
    -- Run-length encoding

end architecture basic ;

